(**************************************************************************
* Useful General-Purpose Tactics for Coq                                  *
* Arthur Charguéraud                                                      *
* Distributed under the terms of the LGPL-v3 license                      *
* Intended for use with Coq version 8.1pl4                                *
***************************************************************************)

(** This file contains a set of tactics that extend the set of builtin
    tactics provided with the standard distribution of Coq. It is 
    intended to overcome a number of limitations with the builtin tactics.
    Hopefully, Coq tactics will be improved in the future and this file
    will no longer be needed. In the meanwhile, these tactics appear
    extremely useful to densify and beautify proof scripts. 
*)

(** The main features offered are:
  - More convenient naming of hypotheses: introduction and inversion 
    need only name hypotheses that are proposition, not necessarily
    all variables. It helps making script more robust without requiring
    a tedious naming of all variables.
  - Improved inversion tactic which substitutes equalities on variables
    generated by the standard inversion mecanism. Moreover, it supports
    the elimination of dependently-typed equalities (requires axiom [K],
    which is a weak form of Proof Irrelevance).
  - Support for manipulating N-ary conjunctions, disjunctions and 
    existentials, hidding the fact that the underlying implementation
    is based on binary predicates.
  - Convenient support for automation: tactic followed with the symbol
    "~" or "*" will call automation on the subgoals they generate.  
    Symbol "~" stands for [auto] and "*" for ([intuition eauto]),  
    but these bindings can be customized.
  - An enhanced implementation of tactic [apply] is provided: 
    it behaves better with respect to conversion.
    (It is not yet clear how this tactic compares with [apply] from V8.2.)
  - A forward-chaining tactic is provided: it can be used to instantiate
    a lemma on existential variables (evars). Often, it avoids 
    providing explicitly all the arguments of a lemma.
  - Finally, we provide with tactics for helping out in the process of
    completing proofs with an improved version of tactic [clear],
    with a tactic [skip] that can be used to temporarily skip any subgoal,
    and with a tactic [sort] that puts all the propositions at the botton
    of the proof context.
*)

(** External credits:
  - thanks to Xavier Leroy for providing with the idea of tactic [forward],
  - thanks to Georges Gonthier for the implementation trick in [rapply],
  - thanks to Hugo Herbelin for useful feedback on several tactics.
*)


(* ********************************************************************** *)
(** * Additional notations for Coq *)

(* ---------------------------------------------------------------------- *)
(** ** N-ary Existentials *)

(** [exist T1 ... TN, P] is a shorthand for 
    [exists T1, ..., exists TN, P]. *)

Notation "'exist' x1 ',' P" :=
  (exists x1, P)
  (at level 200, x1 ident, 
   right associativity) : type_scope. 
Notation "'exist' x1 x2 ',' P" :=
  (exists x1, exists x2, P)
  (at level 200, x1 ident, x2 ident, 
   right associativity) : type_scope.  
Notation "'exist' x1 x2 x3 ',' P" :=
  (exists x1, exists x2, exists x3, P)
  (at level 200, x1 ident, x2 ident, x3 ident, 
   right associativity) : type_scope. 
Notation "'exist' x1 x2 x3 x4 ',' P" :=
  (exists x1, exists x2, exists x3, exists x4, P)
  (at level 200, x1 ident, x2 ident, x3 ident, x4 ident, 
   right associativity) : type_scope.  
Notation "'exist' x1 x2 x3 x4 x5 ',' P" :=
  (exists x1, exists x2, exists x3, exists x4, exists x5, P)
  (at level 200, x1 ident, x2 ident, x3 ident, x4 ident, x5 ident,
   right associativity) : type_scope.  


(* ---------------------------------------------------------------------- *)
(**  ** Partial application of equality *)

(** [= x] is a unary predicate which holds of values equal to [x].
    It simply denotes the partial application of equality. *)

Notation "'=' x" := (fun y => y = x) (at level 71).


(* ********************************************************************** *)
(** * Forward and backward chaining *)

(* ---------------------------------------------------------------------- *)
(**  ** Assertions *)

(** [false] replaces any goal by the goal [False].
    Furthermore, it discharges the obligation if the context contains 
    [False] or an hypothesis of the form [C x1 .. xN  =  D y1 .. yM]. *)

Tactic Notation "false" :=
  assert False; [ try assumption; try discriminate | contradiction ].

(** [tryfalse] tries to solve a goal by contradiction.
    It is equivalent to [try solve \[ false \]]. *)

Tactic Notation "tryfalse" :=
  try solve [ false ].

(** [false T] proves a goal if the term [T] has type [False]
    or if is an equality of the form [C x1 .. xN  =  D y1 .. yM],
    and works if [false] alone solves the goal *)

Tactic Notation "false" constr(T) :=
  solve [ pose T; false ].

(** [asserts H: T] is another syntax for [assert (H : T)], which
    also works with introduction patterns. For instance, we can write:
    [asserts \[x P\] (exists n, n = 3)], or 
    [asserts \[H|H\] (n = 0 \/ n = 1). *)

Tactic Notation "asserts" simple_intropattern(I) ":" constr(T) :=
  let H := fresh in assert (H : T); 
  [ | generalize H; clear H; intros I ].

(** [asserts H1 .. HN: T] is the same as 
    [asserts \[H1 \[H2 \[.. HN\]\]\]\]: T]. *)

Tactic Notation "asserts" simple_intropattern(I1)
 simple_intropattern(I2) ":" constr(T) :=
  asserts [I1 I2]: T.
Tactic Notation "asserts" simple_intropattern(I1)
 simple_intropattern(I2) simple_intropattern(I3) ":" constr(T) :=
  asserts [I1 [I2 I3]]: T.
Tactic Notation "lets" simple_intropattern(I1)
 simple_intropattern(I2) simple_intropattern(I3)
 simple_intropattern(I4) ":" constr(T) :=
  asserts [I1 [I2 [I3 I4]]]: T.
Tactic Notation "lets" simple_intropattern(I1)
 simple_intropattern(I2) simple_intropattern(I3)
 simple_intropattern(I4) simple_intropattern(I5) ":" constr(T) :=
  asserts [I1 [I2 [I3 [I4 I5]]]]: T.
Tactic Notation "lets" simple_intropattern(I1)
 simple_intropattern(I2) simple_intropattern(I3)
 simple_intropattern(I4) simple_intropattern(I5)
 simple_intropattern(I6) ":" constr(T) :=
  asserts [I1 [I2 [I3 [I4 [I5 I6]]]]]: T.

(** [asserts: T as H] is another syntax for [asserts H: T] *)

Tactic Notation "asserts" ":" constr(T) "as" simple_intropattern(I) :=
  asserts I : T.

(** [asserts: T] is [asserts H: T] with [H] being chosen automatically. *)

Tactic Notation "asserts" ":" constr(T) :=
  let H := fresh in asserts H : T.

(** [cuts H: T] is the same as [asserts H: T] except that the two subgoals
    generated are swapped: the subgoal [T] comes second. Note that contrary
    to [cut], it introduces the hypothesis. *)

Tactic Notation "cuts" simple_intropattern(I) ":" constr(T) :=
  cut (T); [ intros I | idtac ].

(** [cuts: T] is [cuts H: T] with [H] being chosen automatically. *)

Tactic Notation "cuts" ":" constr(T) :=
  let H := fresh in cuts H: T.


(* ---------------------------------------------------------------------- *)
(**  ** Adding assumptions *)

(** [lets H: E] adds an hypothesis [H : T] to the context, where [T] is 
    the type of term [E]. If [H] is an introduction pattern, it will
    destruct [H] according to the pattern. *)

Tactic Notation "lets" simple_intropattern(I) ":" constr(E) :=
  generalize E; intros I.

(** [lets H1 .. HN : E] is the same as 
    [lets \[H1 \[H2 \[.. HN\]\]\]\]:E], and thus equivalent to
    [destruct E as \[H1 \[H2 \[.. HN\]\]\]\]]. *)

Tactic Notation "lets" simple_intropattern(I1)
 simple_intropattern(I2) ":" constr(E) :=
  lets [I1 I2]: E.
Tactic Notation "lets" simple_intropattern(I1)
 simple_intropattern(I2) simple_intropattern(I3) ":" constr(E) :=
  lets [I1 [I2 I3]]: E.
Tactic Notation "lets" simple_intropattern(I1)
 simple_intropattern(I2) simple_intropattern(I3)
 simple_intropattern(I4) ":" constr(E) :=
  lets [I1 [I2 [I3 I4]]]: E.
Tactic Notation "lets" simple_intropattern(I1)
 simple_intropattern(I2) simple_intropattern(I3)
 simple_intropattern(I4) simple_intropattern(I5) ":" constr(E) :=
  lets [I1 [I2 [I3 [I4 I5]]]]: E.
Tactic Notation "lets" simple_intropattern(I1)
 simple_intropattern(I2) simple_intropattern(I3)
 simple_intropattern(I4) simple_intropattern(I5)
 simple_intropattern(I6) ":" constr(E) :=
  lets [I1 [I2 [I3 [I4 [I5 I6]]]]]: E.

(** [lets_simpl H: E] is the same as [lets H: E] excepts that it
    calls [simpl] on the hypothesis H. *)

Tactic Notation "lets_simpl" ident(H) ":" constr(E) :=
  lets H: E; simpl in H.

(** [lets_hnf H: E] is the same as [lets H: E] excepts that it
    calls [hnf] to set the definition in head normal form. *)

Tactic Notation "lets_hnf" ident(H) ":" constr(E) :=
  lets H: E; hnf in H.

(** [lets: E] is equivalent to [lets H: E], only the name [H] is
    automatically chosen by Coq. It is useful to type-check a
    term (like the top-level command [Check]), but also to add
    facts that are going to be used by automation. 
    Syntax [lets: E1 .. EN] is short for [lets: E1; ..; lets: EN]. *)

Tactic Notation "lets" ":" constr(E1) :=
  generalize E1; intro.
Tactic Notation "lets" ":" constr(E1) constr(E2) :=
  lets: E1; lets: E2.
Tactic Notation "lets" ":" constr(E1) constr(E2) constr(E3) :=
  lets: E1; lets: E2; lets: E3.

(** [lets_simpl: E] is the same as [lets_simpl H: E] with 
   the name [H] being choosed automatically. *)

Tactic Notation "lets_simpl" ":" constr(T) :=
  let H := fresh in lets_simpl H: T.

(** [lets_hnf: E] is the same as [lets_hnf H: E] with 
   the name [H] being choosed automatically. *)

Tactic Notation "lets_hnf" ":" constr(T) :=
  let H := fresh in lets_hnf H: T.

(** [put X: E] is a synonymous for [pose (X := E)].
    Other syntaxes are [put: E] and [put E]. *)

Tactic Notation "put" ident(X) ":" constr(E) :=
  pose (X := E).
Tactic Notation "put" ":" constr(E) :=
  let X := fresh "X" in pose (X := E).
Tactic Notation "put" constr(E) :=
  put: E.


(* ---------------------------------------------------------------------- *)
(**  ** Forward Chaining *)

(** Underlying implementation of forward *) 

Ltac forwards_tactic I E :=
  match type of E with
  | ?A -> ?B =>
    let HYPA := fresh in 
    assert (HYPA : A); 
    [| forwards_tactic I (E HYPA); clear HYPA ]
  | forall _:?T,_ => 
    let XVAR := fresh in 
    evar (XVAR : T); 
    let HYPB1 := constr:(E XVAR) in
    let HYPB2 := eval unfold XVAR in HYPB1 in 
    forwards_tactic I HYPB2; subst XVAR
  | _ => lets I: E
  end.

Ltac forwards_depth I E N :=
  match N with
  | 0 => lets I: E
  | S ?N' => 
    match type of E with
    | ?A -> ?B =>
      let HYPA := fresh in 
      assert (HYPA : A); 
      [| forwards_depth I (E HYPA) N'; clear HYPA ]
    | forall _:?T,_ => 
      let XVAR := fresh in 
      evar (XVAR : T); 
      let HYPB1 := constr:(E XVAR) in
      let HYPB2 := eval unfold XVAR in HYPB1 in 
      forwards_depth I HYPB2 N'; subst XVAR
    end
  end.

(** [forwards H: E] applied to an expression [E] of type 
    [forall (x1:P1)...(xN:PN).Q] adds the assumption [H:Q] to the 
    context. The variables [xi] are introduced as uninstanciated
    variables if dependent and as subgoals if non-dependent.
    Note that [H] may be a structured introduction pattern. *)

Tactic Notation "forwards" simple_intropattern(I) ":" constr(E) :=  
  forwards_tactic I E.

(** [forwards: E as I] is another syntax for [forwards I: E] *) 

Tactic Notation "forwards" ":" constr(E) "as" simple_intropattern(I) :=  
  forwards_tactic I E.

(** [forwards: E] is same as [forwards H: E] with
    [H] being chosen automatically. *)

Tactic Notation "forwards" ":" constr(E) :=  
  let H := fresh in forwards H: E.

(** [forwards H: E depth N] is similar to [forwards H E] except
    that it ensures that only the [N] first arguments of the 
    proposition [E] will be instantiated. *)

Tactic Notation "forwards" simple_intropattern(I) ":" constr(E) 
 "depth" constr(N) :=  
  forwards_depth I E N.

(** [forwards: E depth N as H] is another syntax for 
    [forwards H: E depth N]. *) 

Tactic Notation "forwards" ":" constr(E) "depth" constr(N)
 "as" simple_intropattern(I) :=  
  forwards_depth I E N.

(** [forwards: E depth N] is same as [forwards H: E depth N] 
    with [H] being chosen automatically. *)

Tactic Notation "forwards" ":" constr(E) "depth" constr(N) :=  
  let H := fresh in forwards H: E depth N.


(* ---------------------------------------------------------------------- *)
(**  ** Application *)

(** [rapply] is a tactic similar to [eapply] except that it is
    based on the [refine] tactics, and thus is strictly more 
    powerful (at least in theory :). In short, it is able to perform
    on-the-fly conversions when required for arguments to match,
    and it is able to instantiate existentials when required. *)

Tactic Notation "rapply" constr(t) :=
  first 
  [ refine (@t _) 
  | refine (@t _ _) 
  | refine (@t _ _ _)
  | refine (@t _ _ _ _)
  | refine (@t _ _ _ _ _)
  | refine (@t _ _ _ _ _ _)
  | refine (@t _ _ _ _ _ _ _)
  | refine (@t _ _ _ _ _ _ _ _)
  | refine (@t _ _ _ _ _ _ _ _ _)
  | refine (@t _ _ _ _ _ _ _ _ _ _)
  | refine (@t _ _ _ _ _ _ _ _ _ _ _)
  | refine (@t _ _ _ _ _ _ _ _ _ _ _ _)
  | refine (@t _ _ _ _ _ _ _ _ _ _ _ _ _)
  ].

(** The tactics [rapply N T], where [N] is a natural number,
    provides a more efficient way of using [rapply T]. It avoids
    trying out all possible arities, by specifying explicitely
    the arity of function [T]. This version is to be preferred
    for programming intensively-used tactics. *)

Tactic Notation "rapply" "0" constr(t) :=
  refine (@t).
Tactic Notation "rapply" "1" constr(t) :=
  refine (@t _).
Tactic Notation "rapply" "2" constr(t) :=
  refine (@t _ _).
Tactic Notation "rapply" "3" constr(t) :=
  refine (@t _ _ _).
Tactic Notation "rapply" "4" constr(t) :=
  refine (@t _ _ _ _).
Tactic Notation "rapply" "5" constr(t) :=
  refine (@t _ _ _ _ _).
Tactic Notation "rapply" "6" constr(t) :=
  refine (@t _ _ _ _ _ _).
Tactic Notation "rapply" "7" constr(t) :=
  refine (@t _ _ _ _ _ _ _).
Tactic Notation "rapply" "8" constr(t) :=
  refine (@t _ _ _ _ _ _ _ _).
Tactic Notation "rapply" "9" constr(t) :=
  refine (@t _ _ _ _ _ _ _ _ _).
Tactic Notation "rapply" "10" constr(t) :=
  refine (@t _ _ _ _ _ _ _ _ _ _).

(** [fapply] is a version of [apply] based on [forward]. *)

Tactic Notation "fapply" constr(E) := 
  let H := fresh in forwards H: E; 
  first [ apply H | eapply H | hnf; apply H 
        | hnf; eapply H | rapply H ].

(** [sapply] stands for "super apply". It tries
    [apply], [eapply], [rapply] and [fapply],
    plus [hnf;apply] and [hnf;eapply]. *)

Tactic Notation "sapply" constr(H) :=
  first [ apply H | eapply H | hnf; apply H 
        | hnf; eapply H | rapply H | fapply H ].

(** [apply_to H E] transform the type of hypothesis [H] by 
    replacing it by the result of the application of the term 
    [E] to [H]. Intuitively, it is equivalent to [lets H: (E H)]. *)

Tactic Notation "apply_to" hyp(H) constr(E) :=
  let H' := fresh in rename H into H';
  (first [ lets H: (E H') 
         | lets H: (E _ H') 
         | lets H: (E _ _ H') 
         | lets H: (E _ _ _ H') 
         | lets H: (E _ _ _ _ H') 
         | lets H: (E _ _ _ _ _ H') 
         | lets H: (E _ _ _ _ _ _ H') 
         | lets H: (E _ _ _ _ _ _ _ H') 
         | lets H: (E _ _ _ _ _ _ _ _ H') 
         | lets H: (E _ _ _ _ _ _ _ _ _ H') ]
  ); clear H'.

(** [apply_in H E] transform the hypothesis [H] by replacing it
    by the result of the application of [H] to the term [E]
    Intuitively, it is equivalent to [lets H: (H E)]. *)

Tactic Notation "apply_in" hyp(H) constr(E)  :=
  let H' := fresh in rename H into H';
  (first [ lets H: (H' E) 
         | lets H: (H' _ E) 
         | lets H: (H' _ _ E) 
         | lets H: (H' _ _ _ E) 
         | lets H: (H' _ _ _ _ E) 
         | lets H: (H' _ _ _ _ _ E) 
         | lets H: (H' _ _ _ _ _ _ E) 
         | lets H: (H' _ _ _ _ _ _ _ E) 
         | lets H: (H' _ _ _ _ _ _ _ _ E) 
         | lets H: (H' _ _ _ _ _ _ _ _ _ E) ]
  ); clear H'.


(* ********************************************************************** *)
(** * Introduction and generalization *)

(* ---------------------------------------------------------------------- *)
(**  ** Introduction *)

(** [introv] iterates [intro] on all universally-quantified "variables"
    at the head of the goal. More precisely, it introduces only the 
    dependently-used variables. For example, [introv] applied
    to the goal [forall x y, P x -> Q] introduces [x] and [y] but not
    the hypothesis [P x]. If the goal is a definition, then it will
    unfold this definition. *)
    
Tactic Notation "introv" :=
  let rec go _ := match goal with
  | |- ?P -> ?Q => idtac
  | |- forall _, _ => intro; try go tt
  end in first [ go tt | intro; go tt ].

  (* Note: this tactic could also be implemented using as follows: 
      Tactic Notation "introv" := 
        first [ progress (intros until 0) 
              | intro; intros until 0 ].
  *) 

(** [introv H1 .. HN] calls [introv] then introduces an hypothesis as [H1],
    then call [introv] again and introduces an hypothesis as [H2], 
    and so on.
    It provides a convenient way of introducing all the arguments of a
    theorem and name only the non-dependent hypotheses. *)

Tactic Notation "introv" simple_intropattern(I) :=
  introv; intros I.
Tactic Notation "introv" simple_intropattern(I1) simple_intropattern(I2) :=
  introv; intros I1; introv I2.
Tactic Notation "introv" simple_intropattern(I1) simple_intropattern(I2) 
 simple_intropattern(I3) :=
  introv; intros I1; introv I2 I3.
Tactic Notation "introv" simple_intropattern(I1) simple_intropattern(I2) 
 simple_intropattern(I3) simple_intropattern(I4) :=
  introv; intros I1; introv I2 I3 I4.
Tactic Notation "introv" simple_intropattern(I1) simple_intropattern(I2) 
 simple_intropattern(I3) simple_intropattern(I4) simple_intropattern(I5) :=
  introv; intros I1; introv I2 I3 I4 I5.

(** [intros_all] repeats [intro] as long as possible. Contrary to [intros],
    it unfolds any definition on the way. Remark that it also unfolds the
    definition of negation, so applying [introz] to a goal of the form
    [forall x, P x -> ~Q] will introduce [x] and [P x] and [Q], and will
    leave [False] in the goal. *)

Tactic Notation "intros_all" :=
  repeat intro.


(* ---------------------------------------------------------------------- *)
(**  ** Generalization *)

(** [gen X1 .. XN] is a shorthand for calling [generalize dependent] 
    successively on variables [XN]...[X1]. Note that the variables
    are generalized in reverse order, following the convention of
    the [generalize] tactic: it means that [X1] will be the first
    quantified variable in the resulting goal. *) 

Tactic Notation "gen" ident(X1) :=
  generalize dependent X1.
Tactic Notation "gen" ident(X1) ident(X2) :=
  gen X2; gen X1.
Tactic Notation "gen" ident(X1) ident(X2) ident(X3) :=
  gen X3; gen X2; gen X1.
Tactic Notation "gen" ident(X1) ident(X2) ident(X3) ident(X4)  :=
  gen X4; gen X3; gen X2; gen X1.
Tactic Notation "gen" ident(X1) ident(X2) ident(X3) ident(X4) ident(X5) :=
  gen X5; gen X4; gen X3; gen X2; gen X1.

(** [generalizes X] is a shorthand for calling [generalize X; clear X]. 
    It is weaker than tactic [gen X] since it does not support
    dependencies. It is mainly intended for writing tactics. *) 
 
Tactic Notation "generalizes" hyp(X) :=
  generalize X; clear X.


(* ---------------------------------------------------------------------- *)
(** ** Naming *)

(** [sets X: E] is the same as [set (X := E) in *], that is,
    it replaces all occurences of [E] by a fresh meta-variable [X]
    whose definition is [E]. *)

Tactic Notation "sets" ident(X) ":" constr(E) :=
  set (X := E) in *.

(** [set_eq X HX: E] generates the equality [X = E],
    for a fresh name [X]. Syntaxes [set_eq_nosubst X: E] 
    and [set_eq_nosubst: E] are also available. *)

Ltac set_eq_base E X HX :=
  set (X := E) in |-;
  (assert (HX : X = E) by reflexivity; clearbody X).

Tactic Notation "set_eq" ident(X) ident(HX) ":" constr(E) :=
  set_eq_base E X HX.
Tactic Notation "set_eq" ident(X) ":" constr(E) :=
  let HX := fresh "EQ" X in set_eq X HX: E.
Tactic Notation "set_eq" ":" constr(E) :=
  let X := fresh "X" in set_eq X: E.

(** [sets_eq X HX: E] replaces [E] by [X] in the current goal and
    generates the equality [X = E], for a fresh name [X].
    Syntaxes [sets_eq X: E] and [sets_eq: E] are also available. *)

Ltac sets_eq_base E X HX :=
  set (X := E) in *;
  (assert (HX : X = E) by reflexivity; clearbody X).

Tactic Notation "sets_eq" ident(X) ident(HX) ":" constr(E) :=
  sets_eq_base E X HX.
Tactic Notation "sets_eq" ident(X) ":" constr(E) :=
  let HX := fresh "EQ" X in sets_eq X HX: E.
Tactic Notation "sets_eq" ":" constr(E) :=
  let X := fresh "X" in sets_eq X: E.


(* ********************************************************************** *)
(** * Rewriting *)

(* ---------------------------------------------------------------------- *)
(**  ** Unfolding *)

(** [unfolds] unfolds the head definition in the goal, i.e. if the
    goal has form [P x1 ... xN] then it calls [unfold P]. *)

Ltac unfolds_base :=
  match goal with
  | |- ?P => unfold P
  | |- ?P _ => unfold P
  | |- ?P _ _ => unfold P
  | |- ?P _ _ _ => unfold P
  | |- ?P _ _ _ _ => unfold P
  | |- ?P _ _ _ _ _ => unfold P
  | |- ?P _ _ _ _ _ _ => unfold P
  | |- ?P _ _ _ _ _ _ _ => unfold P
  | |- ?P _ _ _ _ _ _ _ _ => unfold P
  | _ => intro; unfolds_base
  end. 

Tactic Notation "unfolds" :=
  unfolds_base.

(** [unfolds P1,..,PN] is a shortcut for [unfold P1,..,PN in *]. *)

Tactic Notation "unfolds" reference(F1) :=
  unfold F1 in *.
Tactic Notation "unfolds" reference(F1) "," reference(F2) :=
  unfold F1,F2 in *.
Tactic Notation "unfolds" reference(F1) "," reference(F2) 
 "," reference(F3) :=
  unfold F1,F2,F3 in *.
Tactic Notation "unfolds" reference(F1) "," reference(F2) 
 "," reference(F3) "," reference(F4) :=
  unfold F1,F2,F3,F4 in *.

(** [folds P1,..,PN] is a shortcut for [fold P1 in *; ..; fold PN in *]. *)

Tactic Notation "folds" constr(H) :=
  fold H in *.
Tactic Notation "folds" constr(H1) "," constr(H2) :=
  folds H1; folds H2.
Tactic Notation "folds" constr(H1) "," constr(H2) "," constr(H3) :=
  folds H1; folds H2; folds H3.
Tactic Notation "folds" constr(H1) "," constr(H2) "," constr(H3)
 "," constr(H4) :=
  folds H1; folds H2; folds H3; folds H4.
Tactic Notation "folds" constr(H1) "," constr(H2) "," constr(H3)
 "," constr(H4) "," constr(H5) :=
  folds H1; folds H2; folds H3; folds H4; folds H5.


(* ---------------------------------------------------------------------- *)
(**  ** Simplification *)

(** [simpls] is a shortcut for [simpl in *]. *)

Tactic Notation "simpls" :=
  simpl in *.

(** [simpls P1,..,PN] is a shortcut for 
    [simpl P1 in *; ..; simpl PN in *]. *)

Tactic Notation "simpls" reference(F1) :=
  simpl F1 in *.
Tactic Notation "simpls" reference(F1) "," reference(F2) :=
  simpls F1; simpls F2.
Tactic Notation "simpls" reference(F1) "," reference(F2) 
 "," reference(F3) :=
  simpls F1; simpls F2; simpls F3.
Tactic Notation "simpls" reference(F1) "," reference(F2) 
 "," reference(F3) "," reference(F4) :=
  simpls F1; simpls F2; simpls F3; simpls F4.

(** [unsimpl E] replaces all occurence of [X] by [E], where [X] is 
   the result which the tactic [simpl] would give when applied to [E]. 
   It is useful to undo what [simpl] has simplified too far. *)

Tactic Notation "unsimpl" constr(E) := 
  let F := (eval simpl in E) in change F with E.

(** [unsimpl E in H] is similar to [unsimpl E] but it applies
    inside a particular hypothesis [H]. *)

Tactic Notation "unsimpl" constr(E) "in" hyp(H) := 
  let F := (eval simpl in E) in change F with E in H.

(** [unsimpl E in *] applies [unsimpl E] everywhere possible. *)

Tactic Notation "unsimpl" constr(E) "in" "*" := 
  let F := (eval simpl in E) in change F with E in *.


(* ---------------------------------------------------------------------- *)
(**  ** Substitution *)

(** [substs] does the same as [subst], except that it does not fail
    when there are circular equalities in the context. *)

Tactic Notation "substs" :=
  repeat (match goal with H: ?x = ?y |- _ => 
            first [ subst x | subst y ] end).

(** Implementation of [substs below] *)

Ltac substs_below mark := 
  match goal with H: ?T |- _ =>
  match T with
  | mark => idtac
  | ?x = ?y => 
    first [ subst x; substs_below mark
          | subst y; substs_below mark
          | generalizes H; substs_below mark; intro ]  
  end end.

(** [substs below E] applies [subst] on all equalities that appear
    in the context below the first hypothesis whose body is [E].
    If there is no such hypothesis in the context, it is equivalent
    to [subst]. For instance, if [H] is an hypothesis, then 
    [substs below H] will substitute equalities below hypothesis [H]. *)

Tactic Notation "substs" "below" "body" constr(M) :=
  substs_below M.

(** [substs below H] applies [subst] on all equalities that appear
    in the context below the hypothesis named [H]. Note that 
    the current implementation is technically incorrect since it
    will confuse different hypotheses with the same body. *)

Tactic Notation "substs" "below" hyp(H) :=
  match type of H with ?M => substs below body M end.



(* ---------------------------------------------------------------------- *)
(**  ** Rewriting *)

(** [rewrite_all E] iterates version of [rewrite E] as long as possible. 
    Warning: this tactic can easily get into an infinite loop. 
    Syntax for rewriting from right to left and/or into an hypothese
    is similar to the one of [rewrite]. *)

Tactic Notation "rewrite_all" constr(E) :=
  repeat rewrite E.
Tactic Notation "rewrite_all" "<-" constr(E) :=
  repeat rewrite <- E.
Tactic Notation "rewrite_all" constr(E) "in" ident(H) :=
  repeat rewrite E in H.
Tactic Notation "rewrite_all" "<-" constr(E) "in" ident(H) :=
  repeat rewrite <- E in H.

(** [asserts_rewrite E] asserts that an equality [E] holds (generating a 
    corresponding subgoal) and rewrite it straight away in the current 
    goal. It avoids giving a name to the equality and later clearing it. 
    Syntax for rewriting from right to left and/or into an hypothese
    is similar to the one of [rewrite]. *)

Ltac asserts_rewrite_tactic E action :=
  let EQ := fresh in (assert (EQ : E);
  [ idtac | action EQ; clear EQ ]).

Tactic Notation "asserts_rewrite" constr(E) :=
  asserts_rewrite_tactic E ltac:(fun EQ => rewrite EQ).
Tactic Notation "asserts_rewrite" "<-" constr(E) :=
  asserts_rewrite_tactic E ltac:(fun EQ => rewrite <- EQ).
Tactic Notation "asserts_rewrite" constr(E) "in" hyp(H) :=
  asserts_rewrite_tactic E ltac:(fun EQ => rewrite EQ in H).
Tactic Notation "asserts_rewrite" "<-" constr(E) "in" hyp(H) :=
  asserts_rewrite_tactic E ltac:(fun EQ => rewrite <- EQ in H).

(** [cuts_rewrite E] is the same as [asserts_rewrite E] except
    that subgoals are permuted. *)

Ltac cuts_rewrite_tactic E action :=
  let EQ := fresh in (cuts EQ: E; 
  [ action EQ; clear EQ | idtac ]).

Tactic Notation "cuts_rewrite" constr(E) :=
  cuts_rewrite_tactic E ltac:(fun EQ => rewrite EQ).
Tactic Notation "cuts_rewrite" "<-" constr(E) :=
  cuts_rewrite_tactic E ltac:(fun EQ => rewrite <- EQ).
Tactic Notation "cuts_rewrite" constr(E) "in" hyp(H) :=
  cuts_rewrite_tactic E ltac:(fun EQ => rewrite EQ in H).
Tactic Notation "cuts_rewrite" "<-" constr(E) "in" hyp(H) :=
  cuts_rewrite_tactic E ltac:(fun EQ => rewrite <- EQ in H).

(** [rewrite_except H EQ] rewrites equality [EQ] everywhere
    but in hypothesis [H]. *)

Ltac rewrite_except H EQ :=
  let K := fresh in let T := type of H in 
  set (K := T) in H;
  rewrite EQ in *; unfold K in H; clear K.


(* ---------------------------------------------------------------------- *)
(**  ** Proving equalities *)

(** [fequals] is the same as [f_equal] except that is 
    calls [reflexivity] and [congruence] on all subgoals.
    It applies to goals of the form [f x1 .. xN = f y1 .. yN]
    and tries to solve subgoals [xi = yi], and leaves the
    remaining subgoals. *)

Tactic Notation "fequals" :=
  f_equal; first [ reflexivity | congruence | idtac ].


(* ********************************************************************** *)
(** * Inversion *)

(* ---------------------------------------------------------------------- *)
(**  ** Tools for naming hypotheses generated by inversion *)

(** [mark_type] and [mark] are dummy definitions used as sentinel
    by tactics. The [mark] has type [mark_type]. *)

Inductive mark_type : Type :=
  | mark : mark_type.

(** [gen_until_mark] repeats [generalize] on hypotheses from the 
    context, starting from the bottom and stopping as soon as reaching
    an hypothesis of type [mark_type]. If fails if [mark_type] does not
    appear in the context. *)

Ltac gen_until_mark :=
  match goal with H: ?T |- _ =>
  match T with 
  | mark_type => clear H
  | _ => generalizes H; gen_until_mark
  end end.

(** [intro_until_mark] repeats [intro] until reaching an hypothesis of
    type [mark_type]. It throws away the hypothesis [mark_type]. 
    It fails if [mark_type] does not appear as an hypothesis in the 
    goal. *)

Ltac intro_until_mark :=
  match goal with 
  | |- (mark_type -> _) => intros _
  | _ => intro; intro_until_mark
  end.

(* ---------------------------------------------------------------------- *)
(**  ** Basic inversion *)

(** [invert keep H] is same to [inversion H] except that it puts all the
    facts obtained in the goal. The keyword [keep] means that the
    hypothesis [H] should not be removed. *)

Tactic Notation "invert" "keep" hyp(H) :=
  pose mark; inversion H; gen_until_mark.

(** [invert keep H as X1 .. XN] is the same as [inversion H as ...] except
    that only hypotheses which are not variable need to be named
    explicitely, in a similar fashion as [introv] is used to name
    only hypotheses. *)

Tactic Notation "invert" "keep" hyp(H) "as" simple_intropattern(I1) :=
  invert keep H; introv I1. 
Tactic Notation "invert" "keep" hyp(H) "as" simple_intropattern(I1) 
 simple_intropattern(I2) :=
  invert keep H; introv I1 I2. 
Tactic Notation "invert" "keep" hyp(H) "as" simple_intropattern(I1) 
 simple_intropattern(I2) simple_intropattern(I3) :=
  invert keep H; introv I1 I2 I3. 

(** [invert H] is same to [inversion H] except that it puts all the
    facts obtained in the goal and clears hypothesis [H].
    In other words, it is equivalent to [invert keep H; clear H]. *)

Tactic Notation "invert" hyp(H) :=
  invert keep H; clear H.

(** [invert H as X1 .. XN] is the same as [invert keep H as X1 .. XN]
    but it also clears hypothesis [H]. *)

Tactic Notation "invert_tactic" hyp(H) tactic(tac) :=
  let H' := fresh in rename H into H'; tac H'; clear H'.
Tactic Notation "invert" hyp(H) "as" simple_intropattern(I1) :=
  invert_tactic H (fun H => invert keep H as I1).
Tactic Notation "invert" hyp(H) "as" simple_intropattern(I1) 
 simple_intropattern(I2) :=
  invert_tactic H (fun H => invert keep H as I1 I2).
Tactic Notation "invert" hyp(H) "as" simple_intropattern(I1) 
 simple_intropattern(I2) simple_intropattern(I3) :=
  invert_tactic H (fun H => invert keep H as I1 I2 I3).


(* ---------------------------------------------------------------------- *)
(**  ** Inversion with substitution *)

(** Our inversion tactics is able to get rid of dependent equalities
    generated by [inversion], using proof irrelevance. *)

Require Import Eqdep.

(* todo = EqdepFacts.inj_pair2 *)
Lemma inversion_existT : forall (U:Type) (P:U -> Type) (p:U) (x y:P p),
    existT P p x = existT P p y -> x = y.
Proof.
  intros. lets K: (eq_sigT_eq_dep _ _ _ _ _ _ H).
  exact (eq_dep_eq _ _ _ _ _ K).
Qed.

(* Implicit Arguments inversion_existT [U P p x y]. *)

Ltac inverts_tactic H i1 i2 i3 :=
  let rec go i1 i2 i3 :=
    match goal with 
    | |- (mark_type -> _) => intros _
    | |- (?x = ?y -> _) => let H := fresh in intro H; 
                           first [ subst x | subst y ]; 
                           go i1 i2 i3
    | |- (existT ?P ?p ?x = existT ?P ?p ?y -> _) =>
         let H := fresh in intro H; 
         generalize (@inversion_existT _ P p x y H);
         clear H; go i1 i2 i3
    | |- (?P -> ?Q) => i1; go i2 i3 ltac:(intro)
    | |- (forall _, _) => intro; go i1 i2 i3
    end in
  generalize mark; invert keep H; go i1 i2 i3.

(** [inverts keep H] is same to [invert keep H] except that it 
    applies [subst] to all the equalities generated by the inversion. *)

Tactic Notation "inverts" "keep" hyp(H) :=
  inverts_tactic H ltac:(intro) ltac:(intro) ltac:(intro).

(** [inverts keep H as X1 .. XN] is the same as
    [invert keep H as X1 .. XN] except that it applies [subst] to all the 
    equalities generated by the inversion *)

Tactic Notation "inverts" "keep" hyp(H) "as" simple_intropattern(I1) :=
  inverts_tactic H ltac:(intros I1) ltac:(intro) ltac:(intro).
Tactic Notation "inverts" "keep" hyp(H) "as" simple_intropattern(I1) 
 simple_intropattern(I2) :=
  inverts_tactic H ltac:(intros I1) ltac:(intros I2) ltac:(intro).
Tactic Notation "inverts" "keep" hyp(H) "as" simple_intropattern(I1) 
 simple_intropattern(I2) simple_intropattern(I3) :=
  inverts_tactic H ltac:(intros I1) ltac:(intros I2) ltac:(intros I3).

(** [inverts H] is same to [inverts keep H] except that it 
    clears hypothesis [H]. *)

Tactic Notation "inverts" hyp(H) :=
  inverts keep H; clear H.

(** [inverts H as X1 .. XN] is the same as [inverts keep H as X1 .. XN]
    but it also clears the hypothesis [H]. *)

Tactic Notation "inverts_tactic" hyp(H) tactic(tac) :=
  let H' := fresh in rename H into H'; tac H'; clear H'.
Tactic Notation "inverts" hyp(H) "as" simple_intropattern(I1) :=
  invert_tactic H (fun H => inverts keep H as I1).
Tactic Notation "inverts" hyp(H) "as" simple_intropattern(I1) 
 simple_intropattern(I2) :=
  invert_tactic H (fun H => inverts keep H as I1 I2).
Tactic Notation "inverts" hyp(H) "as" simple_intropattern(I1) 
 simple_intropattern(I2) simple_intropattern(I3) :=
  invert_tactic H (fun H => inverts keep H as I1 I2 I3).


(* ---------------------------------------------------------------------- *)
(**  ** Injection with substitution *)

(** Underlying implementation of [injects] *)

Ltac injects_tactic H :=
  let rec go _ :=
    match goal with 
    | |- (mark_type -> _) => intros _
    | |- (?x = ?y -> _) => let H := fresh in intro H; 
                           first [ subst x | subst y | idtac ]; 
                           go tt
    end in
  generalize mark; injection H; go tt.

(** [injects keep H] takes an hypothesis [H] of the form
    [C a1 .. aN = C b1 .. bN] and substitute all equalities
    [ai = bi] that have been generated. *)

Tactic Notation "injects" "keep" hyp(H) :=
  injects_tactic H.

(** [injects H] is similar to [injects keep H] but clears 
    the hypothesis [H]. *)

Tactic Notation "injects" hyp(H) :=
  injects_tactic H; clear H.

(** [inject H as X1 .. XN] is the same as [injection]
    followed by [intros X1 .. XN] *)
   
Tactic Notation "inject" hyp(H) :=
  injection H. 
Tactic Notation "inject" hyp(H) "as" ident(X1) :=
  injection H; intros X1.
Tactic Notation "inject" hyp(H) "as" ident(X1) ident(X2) :=
  injection H; intros X1 X2.
Tactic Notation "inject" hyp(H) "as" ident(X1) ident(X2) ident(X3) :=
  injection H; intros X1 X2 X3.
Tactic Notation "inject" hyp(H) "as" ident(X1) ident(X2) ident(X3) 
 ident(X4) :=
  injection H; intros X1 X2 X3 X4.
Tactic Notation "inject" hyp(H) "as" ident(X1) ident(X2) ident(X3) 
 ident(X4) ident(X5) :=
  injection H; intros X1 X2 X3 X4 X5.


(* ---------------------------------------------------------------------- *)
(**  ** Inversion and injection with substitution --rough implementation *)

(** The tactics [inversions] and [injections] provided in this section
    are similar to [inverts] and [injects] except that they perform
    substitution on all equalities from the context and not only
    the ones freshly generated. The counterpart is that they have
    simpler implementations. *)

(** [inversions keep H] is the same as [inversions H] but it does
    not clear hypothesis [H]. *)

Tactic Notation "inversions" "keep" hyp(H) :=
  inversion H; subst.

(** [inversions H] is a shortcut for [inversion H] followed by [subst]
    and [clear H].
    It is a rough implementation of [inverts keep H] which behave
    badly when the proof context already contains equalities. 
    It is provided in case the better implementation turns out to be
    too slow. *)

Tactic Notation "inversions" hyp(H) :=
  inversion H; subst; clear H.

(** [injections keep H] is the same as [injection H] followed
    by [intros] and [subst]. It is a rough implementation of 
    [injects keep H] which behave
    badly when the proof context already contains equalities,
    or when the goal starts with a forall or an implication. *)

Tactic Notation "injections" "keep" hyp(H) :=
  injection H; intros; subst.

(** [injections H] is the same as [injection H] followed
    by [intros] and [clear H] and [subst]. It is a rough 
    implementation of [injects keep H] which behave
    badly when the proof context already contains equalities,
    or when the goal starts with a forall or an implication. *)

Tactic Notation "injections" "keep" hyp(H) :=
  injection H; clear H; intros; subst.


(* ********************************************************************** *)
(** * Induction *)

(** Induction has a well-known limitation: it "forgets" information
    when the inductive proposition being eliminated mentions terms
    that are not reduced to variables. The work-around which we use
    follows three steps:
    - assign a name to the terms which are not variables,
      and generalize an equality binding these names to terms;
      (use [gen_eq] defined further).
    - apply the [induction] tactic;
    - substitute for the temporary names introduced ([subst]).
 *)

(** [gen_eq X: E] is a tactic whose purpose is to introduce 
    equalities so as to work around the limitation of the [induction]
    tactic which typically loses information. [gen_eq E as X] replaces 
    all occurences of term [E] with a fresh variable [X] and the equality
    [X = E] as extra hypothesis to the current conclusion. In other words
    a conclusion [C] will be turned into [(X = E) -> C].
    [gen_eq: E] and [gen_eq: E as X] are also accepted. *)

Tactic Notation "gen_eq" ident(X) ":" constr(E) :=
  let EQ := fresh in sets_eq X EQ: E; revert EQ.
Tactic Notation "gen_eq" ":" constr(E) :=
  let X := fresh "X" in gen_eq X: E.
Tactic Notation "gen_eq" ":" constr(E) "as" ident(X) :=
  gen_eq X: E.

(** [induction_wf IH: E X] is used to apply the well-founded induction
    principle, for a given well-founded relation. It applies to a goal
    [PX] where [PX] is a proposition on [X]. First, it sets up the
    goal in the form [(fun a => P a) X], using [pattern X], and then
    it applies the well-founded induction principle instantiated on [E].
    Syntaxes [induction_wf: E X] and [induction_wf E X]
    and [induction_wf E X as IH] are also available. *)

Tactic Notation "induction_wf" ident(IH) ":" constr(E) ident(X) :=
  pattern X; apply (well_founded_ind E); clear X; intros X IH.
Tactic Notation "induction_wf" ":" constr(E) ident(X) :=
  let IH := fresh "IH" in induction_wf IH: E X.
Tactic Notation "induction_wf" ":" constr(E) ident(X) "as" ident(IH) :=
  induction_wf IH: E X.
Tactic Notation "induction_wf" ":" constr(E) ident(X) :=
  induction_wf: E X.


(* ********************************************************************** *)
(** * Decidable equality *)

(** [decides_equality] is the same as [decide equality] excepts that it 
    is able to unfold definitions at head of the current goal. *)

Ltac decides_equality_tactic :=
  first [ decide equality | progress(unfolds); decides_equality_tactic ].

Tactic Notation "decides_equality" :=
  decides_equality_tactic.


(* ********************************************************************** *)
(** * N-ary Conjunctions and Disjunctions *)

(* ---------------------------------------------------------------------- *)
(** N-ary Conjunctions Splitting in Goals *)

(** Underlying implementation of [splits]. *)

Ltac splits_tactic N :=
  match N with
  | O => fail
  | S O => idtac
  | S ?N' => split; [| splits_tactic N']
  end.

Ltac unfold_goal_until_conjunction :=
  match goal with
  | |- _ /\ _ => idtac
  | _ => progress(unfolds); unfold_goal_until_conjunction
  end.

Ltac get_term_conjunction_arity T :=
  match T with
  | _ /\ _ /\ _ /\ _ /\ _ /\ _ /\ _ /\ _ => constr:(8)
  | _ /\ _ /\ _ /\ _ /\ _ /\ _ /\ _ => constr:(7)
  | _ /\ _ /\ _ /\ _ /\ _ /\ _ => constr:(6)
  | _ /\ _ /\ _ /\ _ /\ _ => constr:(5)
  | _ /\ _ /\ _ /\ _ => constr:(4)
  | _ /\ _ /\ _ => constr:(3)
  | _ /\ _ => constr:(2)
  | _ -> ?T' => get_term_conjunction_arity T'
  | _ => progress (hnf); get_term_conjunction_arity T
         (* todo: check it does not loop *)
  end.

Ltac get_goal_conjunction_arity :=
  match goal with |- ?T => get_term_conjunction_arity T end.
  
(** [splits] applies to a goal of the form [(T1 /\ .. /\ TN)] and
    destruct it into [N] subgoals [T1] .. [TN]. If the goal is not a 
    conjunction, then it unfolds the head definition. *)

Tactic Notation "splits" :=
  unfold_goal_until_conjunction;
  let N := get_goal_conjunction_arity in
  splits_tactic N.

(** [splits N] is similar to [splits], except that it will unfold as many
    definitions as necessary to obtain an [N]-ary conjunction. *)

Tactic Notation "splits" constr(N) := 
  splits_tactic N.


(* ---------------------------------------------------------------------- *)
(** N-ary Conjunctions Deconstruction *)

(** Underlying implementation of [destructs]. *)

Ltac destructs_conjunction_tactic N T :=
  match N with
  | 2 => destruct T as [? ?]
  | 3 => destruct T as [? [? ?]]
  | 4 => destruct T as [? [? [? ?]]]
  | 5 => destruct T as [? [? [? [? ?]]]]
  | 6 => destruct T as [? [? [? [? [? ?]]]]]
  | 7 => destruct T as [? [? [? [? [? [? ?]]]]]]
  end. (* todo: check there is no bug with "?" in patterns *)

(** [destructs T] allows destructing a term [T] which is a N-ary 
    conjunction. It is equivalent to [destruct T as (H1 .. HN)], 
    except that it does not require to manually specify N different 
    names. *)

Tactic Notation "destructs" constr(T) :=
  let TT := type of T in
  let N := get_term_conjunction_arity TT in 
  destructs_conjunction_tactic N T.

(** [destructs N T] is equivalent to [destruct T as (H1 .. HN)],
    except that it does not require to manually specify N different 
    names. Remark that it is not restricted to N-ary conjunctions. *)

Tactic Notation "destructs" constr(N) constr(T) :=
  destructs_conjunction_tactic N T.


(* ---------------------------------------------------------------------- *)
(** Proving goals which are N-ary disjunctions *)

(** Underlying implementation of [branch]. *)

Ltac branch_tactic M N :=
  match M with
  | O => fail
  | S ?M' => 
    match N with
    | O => fail
    | 1 => left
    | S ?N' => right; branch_tactic M' N'
    end
  end.

Ltac unfold_goal_until_disjunction :=
  match goal with
  | |- _ \/ _ => idtac
  | _ => progress(unfolds); unfold_goal_until_disjunction
  end.

Ltac get_term_disjunction_arity T :=
  match T with
  | _ \/ _ \/ _ \/ _ \/ _ \/ _ \/ _ \/ _ => constr:(8)
  | _ \/ _ \/ _ \/ _ \/ _ \/ _ \/ _ => constr:(7)
  | _ \/ _ \/ _ \/ _ \/ _ \/ _ => constr:(6)
  | _ \/ _ \/ _ \/ _ \/ _ => constr:(5)
  | _ \/ _ \/ _ \/ _ => constr:(4)
  | _ \/ _ \/ _ => constr:(3)
  | _ \/ _ => constr:(2)
  | _ -> ?T' => get_term_disjunction_arity T' (* TEST ?? *)
  end.

Ltac get_goal_disjunction_arity :=
  match goal with |- ?T => get_term_disjunction_arity T end.

(** [branch N] applies to a goal of the form 
    [P1 \/ ... \/ PN \/ ... \/ PM] and leaves the goal [PN].
    It only able to unfold the head definition (if there is one),
    but for more complex unfolding one should use the tactic
    [branch N of M]. *)

Tactic Notation "branch" constr(N) :=
  unfold_goal_until_disjunction; 
  let M := get_goal_disjunction_arity in
  branch_tactic M N.

(** [branch N of M] is similar to [branch N] except that the
    arity of the disjunction [M] is given manually, and so this
    version of the tactic is able to unfold definitions.
    In other words, applies to a goal of the form 
    [P1 \/ ... \/ PN \/ ... \/ PM] and leaves the goal [PK]. *)

Tactic Notation "branch" constr(K) "of" constr(N) :=
  branch_tactic N K.


(* ---------------------------------------------------------------------- *)
(** N-ary Disjunction Deconstruction *)

(** Underlying implementation of [branches]. *)

Ltac destructs_disjunction_tactic N T :=
  match N with
  | 2 => destruct T as [? | ?]
  | 3 => destruct T as [? | ? | ?]
  | 4 => destruct T as [? | ? | ? | ?]
  | 5 => destruct T as [? | ? | ? | ? | ?]
  | 6 => destruct T as [? | ? | ? | ? | ? | ?]
  | 7 => destruct T as [? | ? | ? | ? | ? | ? | ?]
  end.

(** [branches T] allows destructing a term [T] which is a N-ary 
    disjunction. It is equivalent to [destruct T as [ H1 | .. | HN ] ], 
    and produces [N] subgoals corresponding to the [N] possible cases. *)

Tactic Notation "branches" constr(T) :=
  let TT := type of T in
  let N := get_term_disjunction_arity TT in
  destructs_disjunction_tactic N T.

(** [branches N T] is the same as [branches T] except that the arity is
    forced to [N]. This version is useful to unfold definitions 
    on the fly. *)

Tactic Notation "branches" constr(N) constr(T) :=
  destructs_disjunction_tactic T N.
 

(* ---------------------------------------------------------------------- *)
(** N-ary Existentials *)

(** [exists T1 ... TN] is a shorthand for [exists T1; ...; exists TN].
    It is intended to prove goals of the form [exist X1 .. XN, P]. *)

Tactic Notation "exists" constr(T1) :=
  exists T1.
Tactic Notation "exists" constr(T1) constr(T2) :=
  exists T1; exists T2.
Tactic Notation "exists" constr(T1) constr(T2) constr(T3) :=
  exists T1; exists T2; exists T3.
Tactic Notation "exists" constr(T1) constr(T2) constr(T3) constr(T4) :=
  exists T1; exists T2; exists T3; exists T4.
Tactic Notation "exists" constr(T1) constr(T2) constr(T3) constr(T4) 
 constr(T5) :=
  exists T1; exists T2; exists T3; exists T4; exists T5.



(* ********************************************************************** *)
(** * Tactics to invoke automation *)

(* ---------------------------------------------------------------------- *)
(** ** Definitions of automation tactics *)

(** The two following tactics defined the default behaviour of 
    "light automation" and "strong automation". These tactics
    may be redefined at any time using the syntax [Ltac .. ::= ..]. *)

(** [auto_tilde] is the tactic which will be called each time a symbol
    [~] is used after a tactic. *)

Ltac auto_tilde := auto.

(** [auto_star] is the tactic which will be called each time a symbol
    [*] is used after a tactic. *)

Ltac auto_star := try solve [ auto | intuition eauto ].

(** [auto~] is a notation for tactic [auto_tilde]. It may be followed
    by lemmas (or proofs terms) which auto will be able to use
    for solving the goal. *)

Tactic Notation "auto" "~" :=
  auto_tilde.
Tactic Notation "auto" "~" constr(E1) :=
  lets: E1; auto_tilde.
Tactic Notation "auto" "~" constr(E1) constr(E2) :=
  lets: E1 E2; auto_tilde.
Tactic Notation "auto" "~" constr(E1) constr(E2) constr(E3) :=
  lets: E1 E2 E3; auto_tilde.

(** [auto*] is a notation for tactic [auto_star]. It may be followed
    by lemmas (or proofs terms) which auto will be able to use
    for solving the goal. *)

Tactic Notation "auto" "*" :=
  auto_star.
Tactic Notation "auto" "*" constr(E1) :=
  lets: E1; auto_star.
Tactic Notation "auto" "*" constr(E1) constr(E2) :=
  lets: E1 E2; auto_star.
Tactic Notation "auto" "*" constr(E1) constr(E2) constr(E3) :=
  lets: E1 E2 E3; auto_star.


(* ---------------------------------------------------------------------- *)
(** ** Definitions for parsing compatibility *)

Tactic Notation "f_equal" :=
  f_equal.
Tactic Notation "constructor" := 
  constructor.


(* ---------------------------------------------------------------------- *)
(** ** Parsing for light automation *)

(** Any tactic followed by the symbol [~] will have [auto_tilde] called
    on all of its subgoals. Two exceptions:
    - [cuts] and [asserts] only call [auto] on their first subgoal,
    - [apply~] relies on [sapply] rather than [apply]. 

   Some builtin tactics are not defined using tactic notations 
   and thus cannot be extended, e.g. [simpl] and [unfold].
   For these, notation such as [simpl~] will not be available. *)

Tactic Notation "apply" "~" constr(H) :=
  sapply H; auto_tilde.
Tactic Notation "destruct" "~" constr(H) :=
  destruct H; auto_tilde.
Tactic Notation "destruct" "~" constr(H) "as" simple_intropattern(I) :=
  destruct H as I; auto_tilde.
Tactic Notation "f_equal" "~" :=
  f_equal; auto_tilde.
Tactic Notation "induction" "~" constr(H) :=
  induction H; auto_tilde.
Tactic Notation "inversion" "~" constr(H) :=
  inversion H; auto_tilde.
Tactic Notation "split" "~" :=
  split; auto_tilde.
Tactic Notation "subst" "~" :=
  subst; auto_tilde.
Tactic Notation "right" "~" := 
  right; auto_tilde.
Tactic Notation "left" "~" := 
  left; auto_tilde.
Tactic Notation "constructor" "~" := 
  constructor; auto_tilde.

Tactic Notation "false" "~" :=
  false; auto_tilde.
Tactic Notation "asserts" "~" simple_intropattern(H) ":" constr(E) :=
  asserts H: E; [ auto_tilde | idtac ].
Tactic Notation "cuts" "~" simple_intropattern(H) ":" constr(E) :=
  cuts H: E; [ auto_tilde | idtac ].
Tactic Notation "cuts" "~" ":" constr(E) :=
  cuts: E; [ auto_tilde | idtac ].

Tactic Notation "lets" "~" simple_intropattern(I) ":" constr(E) :=
  lets I: E; auto_tilde.
Tactic Notation "lets" "~" ":" constr(E1) :=
  lets: E1; auto_tilde.
Tactic Notation "lets" "~" ":" constr(E1) constr(E2) :=
  lets: E1 E2; auto_tilde.
Tactic Notation "lets" "~" ":" constr(E1) constr(E2) constr(E3) :=
  lets: E1 E2 E3; auto_tilde.

Tactic Notation "forwards" "~" simple_intropattern(I) ":" constr(E) :=  
  forwards I: E; auto_tilde.
Tactic Notation "forwards" "~" ":" constr(E) "as" simple_intropattern(I) :=  
  forwards: E as I; auto_tilde.
Tactic Notation "forwards" "~" ":" constr(E) :=  
  forwards: E; auto_tilde.

Tactic Notation "rapply" "~" constr(E) :=
  rapply E; auto_tilde.
Tactic Notation "fapply" "~" constr(E) := 
  fapply E; auto_tilde.
Tactic Notation "sapply" "~" constr(E) :=
  sapply E; auto_tilde.

Tactic Notation "intros_all" "~" :=
  intros_all; auto_tilde.

Tactic Notation "unfolds" "~" reference(F1) :=
  unfolds F1; auto_tilde.
Tactic Notation "unfolds" "~" reference(F1) reference(F2) :=
  unfolds F1, F2; auto_tilde.
Tactic Notation "unfolds" "~" reference(F1) reference(F2) reference(F3) :=
  unfolds F1, F2, F3; auto_tilde.
Tactic Notation "unfolds" "~" reference(F1) reference(F2) reference(F3) 
 reference(F4) :=
  unfolds F1, F2, F3, F4; auto_tilde.

Tactic Notation "simpls" "~" :=
  simpls; auto_tilde.
Tactic Notation "substs" "~" :=
  substs; auto_tilde.

Tactic Notation "rewrite" "~" constr(E) :=
  rewrite E; auto_tilde.
Tactic Notation "rewrite" "~" "<-" constr(E) :=
  rewrite <- E; auto_tilde.
Tactic Notation "rewrite" "~" constr(E) "in" hyp(H) :=
  rewrite E in H; auto_tilde.
Tactic Notation "rewrite" "~" "<-" constr(E) "in" hyp(H) :=
  rewrite <- E in H; auto_tilde.

Tactic Notation "rewrite_all" "~" constr(E) :=
  rewrite_all E; auto_tilde.
Tactic Notation "rewrite_all" "~" "<-" constr(E) :=
  rewrite_all <- E; auto_tilde.
Tactic Notation "rewrite_all" "~" constr(E) "in" ident(H) :=
  rewrite_all E in H; auto_tilde.
Tactic Notation "rewrite_all" "~" "<-" constr(E) "in" ident(H) :=
  rewrite_all <- E in H; auto_tilde.

Tactic Notation "asserts_rewrite" "~" constr(E) :=
  asserts_rewrite E; auto_tilde.
Tactic Notation "asserts_rewrite" "~" "<-" constr(E) :=
  asserts_rewrite <- E; auto_tilde.
Tactic Notation "asserts_rewrite" "~" constr(E) "in" hyp(H) :=
  asserts_rewrite E in H; auto_tilde.
Tactic Notation "asserts_rewrite" "~" "<-" constr(E) "in" hyp(H) :=
  asserts_rewrite <- E in H; auto_tilde.

Tactic Notation "cuts_rewrite" "~" constr(E) :=
  cuts_rewrite E; auto_tilde.
Tactic Notation "cuts_rewrite" "~" "<-" constr(E) :=
  cuts_rewrite <- E; auto_tilde.
Tactic Notation "cuts_rewrite" "~" constr(E) "in" hyp(H) :=
  cuts_rewrite E in H; auto_tilde.
Tactic Notation "cuts_rewrite" "~" "<-" constr(E) "in" hyp(H) :=
  cuts_rewrite <- E in H; auto_tilde.

Tactic Notation "fequals" "~" :=
  fequals; auto_tilde.

Tactic Notation "invert" "~" hyp(H) :=
  invert H; auto_tilde.
Tactic Notation "inverts" "~" hyp(H) :=
  inverts H; auto_tilde.
Tactic Notation "injects" "~" hyp(H) :=
  injects H; auto_tilde.
Tactic Notation "inversions" "~" hyp(H) :=
  inversions H; auto_tilde.

Tactic Notation "decides_equality" "~" :=
  decides_equality; auto_tilde.

Tactic Notation "splits" "~" :=
  splits; auto_tilde.
Tactic Notation "splits" "~" constr(N) := 
  splits N; auto_tilde.

Tactic Notation "destructs" "~" constr(T) :=
  destructs T; auto_tilde.
Tactic Notation "destructs" "~" constr(N) constr(T) :=
  destructs N T; auto_tilde.

Tactic Notation "branch" "~" constr(N) :=
  branch N; auto_tilde.
Tactic Notation "branch" "~" constr(K) "of" constr(N) :=
  branch K of N; auto_tilde.

Tactic Notation "branches" "~" constr(T) :=
  branches T; auto_tilde.
Tactic Notation "branches" "~" constr(N) constr(T) :=
  branches N T; auto_tilde.

Tactic Notation "exists" "~" constr(T1) :=
  exists T1; auto_tilde.
Tactic Notation "exists" "~" constr(T1) constr(T2) :=
  exists T1 T2; auto_tilde.
Tactic Notation "exists" "~" constr(T1) constr(T2) constr(T3) :=
  exists T1 T2 T3; auto_tilde.
Tactic Notation "exists" "~" constr(T1) constr(T2) constr(T3) constr(T4) :=
  exists T1 T2 T3 T4; auto_tilde.
Tactic Notation "exists" "~" constr(T1) constr(T2) constr(T3) constr(T4) 
 constr(T5) :=
  exists T1 T2 T3 T4 T5; auto_tilde.


(* ---------------------------------------------------------------------- *)
(** ** Parsing for strong automation *)

(** Any tactic followed by the symbol [*] will have [auto*] called
    on all of its subgoals. The exceptions to these rules are the
    same as for light automation. *)

Tactic Notation "apply" "*" constr(H) :=
  sapply H; auto_star.
Tactic Notation "destruct" "*" constr(H) :=
  destruct H; auto_star.
Tactic Notation "destruct" "*" constr(H) "as" simple_intropattern(I) :=
  destruct H as I; auto_star.
Tactic Notation "f_equal" "*" :=
  f_equal; auto_star.
Tactic Notation "induction" "*" constr(H) :=
  induction H; auto_star.
Tactic Notation "inversion" "*" constr(H) :=
  inversion H; auto_star.
Tactic Notation "split" "*" :=
  split; auto_star.
Tactic Notation "subst" "*" :=
  subst; auto_star.
Tactic Notation "right" "*" := 
  right; auto_star.
Tactic Notation "left" "*" := 
  left; auto_star.
Tactic Notation "constructor" "*" := 
  constructor; auto_star.

Tactic Notation "false" "*" :=
  false; auto_star.
Tactic Notation "asserts" "*" simple_intropattern(H) ":" constr(E) :=
  asserts H: E; [ auto_star | idtac ].
Tactic Notation "cuts" "*" simple_intropattern(H) ":" constr(E) :=
  cuts H: E; [ auto_star | idtac ].
Tactic Notation "cuts" "*" ":" constr(E) :=
  cuts: E; [ auto_star | idtac ].

Tactic Notation "lets" "*" simple_intropattern(I) ":" constr(E) :=
  lets I: E; auto_star.
Tactic Notation "lets" "*" ":" constr(E1) :=
  lets: E1; auto_star.
Tactic Notation "lets" "*" ":" constr(E1) constr(E2) :=
  lets: E1 E2; auto_star.
Tactic Notation "lets" "*" ":" constr(E1) constr(E2) constr(E3) :=
  lets: E1 E2 E3; auto_star.

Tactic Notation "forwards" "*" simple_intropattern(I) ":" constr(E) :=  
  forwards I: E; auto_star.
Tactic Notation "forwards" "*" ":" constr(E) "as" simple_intropattern(I) :=  
  forwards: E as I; auto_star.
Tactic Notation "forwards" "*" ":" constr(E) :=  
  forwards: E; auto_star.

Tactic Notation "rapply" "*" constr(E) :=
  rapply E; auto_star.
Tactic Notation "fapply" "*" constr(E) := 
  fapply E; auto_star.
Tactic Notation "sapply" "*" constr(E) :=
  sapply E; auto_star.

Tactic Notation "intros_all" "*" :=
  intros_all; auto_star.

Tactic Notation "unfolds" "*" reference(F1) :=
  unfolds F1; auto_star.
Tactic Notation "unfolds" "*" reference(F1) reference(F2) :=
  unfolds F1, F2; auto_star.
Tactic Notation "unfolds" "*" reference(F1) reference(F2) reference(F3) :=
  unfolds F1, F2, F3; auto_star.
Tactic Notation "unfolds" "*" reference(F1) reference(F2) reference(F3) 
 reference(F4) :=
  unfolds F1, F2, F3, F4; auto_star.

Tactic Notation "simpls" "*" :=
  simpls; auto_star.
Tactic Notation "substs" "*" :=
  substs; auto_star.

Tactic Notation "rewrite" "*" constr(E) :=
  rewrite E; auto_star.
Tactic Notation "rewrite" "*" "<-" constr(E) :=
  rewrite <- E; auto_star.
Tactic Notation "rewrite" "*" constr(E) "in" hyp(H) :=
  rewrite E in H; auto_star.
Tactic Notation "rewrite" "*" "<-" constr(E) "in" hyp(H) :=
  rewrite <- E in H; auto_star.

Tactic Notation "rewrite_all" "*" constr(E) :=
  rewrite_all E; auto_star.
Tactic Notation "rewrite_all" "*" "<-" constr(E) :=
  rewrite_all <- E; auto_star.
Tactic Notation "rewrite_all" "*" constr(E) "in" ident(H) :=
  rewrite_all E in H; auto_star.
Tactic Notation "rewrite_all" "*" "<-" constr(E) "in" ident(H) :=
  rewrite_all <- E in H; auto_star.

Tactic Notation "asserts_rewrite" "*" constr(E) :=
  asserts_rewrite E; auto_star.
Tactic Notation "asserts_rewrite" "*" "<-" constr(E) :=
  asserts_rewrite <- E; auto_star.
Tactic Notation "asserts_rewrite" "*" constr(E) "in" hyp(H) :=
  asserts_rewrite E; auto_star.
Tactic Notation "asserts_rewrite" "*" "<-" constr(E) "in" hyp(H) :=
  asserts_rewrite <- E; auto_star.

Tactic Notation "cuts_rewrite" "*" constr(E) :=
  cuts_rewrite E; auto_star.
Tactic Notation "cuts_rewrite" "*" "<-" constr(E) :=
  cuts_rewrite <- E; auto_star.
Tactic Notation "cuts_rewrite" "*" constr(E) "in" hyp(H) :=
  cuts_rewrite E in H; auto_star.
Tactic Notation "cuts_rewrite" "*" "<-" constr(E) "in" hyp(H) :=
  cuts_rewrite <- E in H; auto_star.

Tactic Notation "fequals" "*" :=
  fequals; auto_star.

Tactic Notation "invert" "*" hyp(H) :=
  invert H; auto_star.
Tactic Notation "inverts" "*" hyp(H) :=
  inverts H; auto_star.
Tactic Notation "injects" "*" hyp(H) :=
  injects H; auto_star.
Tactic Notation "inversions" "*" hyp(H) :=
  inversions H; auto_star.

Tactic Notation "decides_equality" "*" :=
  decides_equality; auto_star.

Tactic Notation "splits" "*" :=
  splits; auto_star.
Tactic Notation "splits" "*" constr(N) := 
  splits N; auto_star.

Tactic Notation "destructs" "*" constr(T) :=
  destructs T; auto_star.
Tactic Notation "destructs" "*" constr(N) constr(T) :=
  destructs N T; auto_star.

Tactic Notation "branch" "*" constr(N) :=
  branch N; auto_star.
Tactic Notation "branch" "*" constr(K) "of" constr(N) :=
  branch K of N; auto_star.

Tactic Notation "branches" "*" constr(T) :=
  branches T; auto_star.
Tactic Notation "branches" "*" constr(N) constr(T) :=
  branches N T; auto_star.

Tactic Notation "exists" "*" constr(T1) :=
  exists T1; auto_star.
Tactic Notation "exists" "*" constr(T1) constr(T2) :=
  exists T1 T2; auto_star.
Tactic Notation "exists" "*" constr(T1) constr(T2) constr(T3) :=
  exists T1 T2 T3; auto_star.
Tactic Notation "exists" "*" constr(T1) constr(T2) constr(T3) constr(T4) :=
  exists T1 T2 T3 T4; auto_star.
Tactic Notation "exists" "*" constr(T1) constr(T2) constr(T3) constr(T4) 
 constr(T5) :=
  exists T1 T2 T3 T4 T5; auto_star.



(* ********************************************************************** *)
(** * Tactics to sort out the proof context *)

(* ---------------------------------------------------------------------- *)
(** ** Sorting hypotheses *)

(** [sort] sorts out hypotheses from the context by moving all the
    propositions (hypotheses of type Prop) to the bottom of the context. *)

Ltac sort_tactic :=
  match goal with H: ?T |- _ =>
  match type of T with Prop =>
    generalizes H; (try sort_tactic); intro
  end end.

Tactic Notation "sort" :=
  sort_tactic.


(* ---------------------------------------------------------------------- *)
(** ** Clearing hypotheses *)

(** [clears X1 ... XN] is a variation on [clear] which clears
    the variables [X1]..[XN] as well as all the hypotheses which
    depend on them. Contrary to [clear], it never fails. *)

Tactic Notation "clears" ident(X1) :=
  let rec doit _ :=
  match goal with 
  | H:context[X1] |- _ => clear H; try (doit tt)
  | _ => clear X1
  end in doit tt.
Tactic Notation "clears" ident(X1) ident(X2) :=
  clears X1; clears X2.
Tactic Notation "clears" ident(X1) ident(X2) ident(X3) :=
  clears X1; clears X2; clear X3.
Tactic Notation "clears" ident(X1) ident(X2) ident(X3) ident(X4) :=
  clears X1; clears X2; clear X3; clear X4.
Tactic Notation "clears" ident(X1) ident(X2) ident(X3) ident(X4) 
 ident(X5) :=
  clears X1; clears X2; clear X3; clear X4; clear X5.
Tactic Notation "clears" ident(X1) ident(X2) ident(X3) ident(X4)
 ident(X5) ident(X6) :=
  clears X1; clears X2; clear X3; clear X4; clear X5; clear X6.

(** [clears] (without any argument) clears all the unused variables
    from the context. In other words, it removes any variable
    which is not a proposition (i.e. not of type Prop) and which 
    does not appear in another hypothesis nor in the goal. *)

Ltac clears_tactic :=
  match goal with H: ?T |- _ =>
  match type of T with 
  | Prop => generalizes H; (try clears_tactic); intro
  | ?TT => clear H; (try clears_tactic)
  | ?TT => generalizes H; (try clears_tactic); intro
  end end.

Tactic Notation "clears" :=
  clears_tactic.


(* ********************************************************************** *)
(** * Tools for programming with Ltac *)

(* ---------------------------------------------------------------------- *)
(** ** Programming tactics  *)

(** [ltac_no_arg] is a constant that can be used to simulate 
    optional arguments in tactic definitions. 
    Use [mytactic ltac_no_arg] on the tactic invokation,
    and use [match arg with ltac_no_arg => ..] or
    [match type of arg with ltac_no_arg_type  => ..] to
    test whether an argument was provided. *)

Inductive ltac_no_arg_type : Set := 
  | ltac_no_arg : ltac_no_arg_type.

(** Ltac tactics are not allowed to both perform side-effect on the goal
    and return a value in the same time. To work around this limitation,
    we can use the proof context as a stack to place return values.
    The type [carrier_type] is used to box return values, so that they
    do not get mixed up with the other hypotheses. *) 

Inductive carrier_type : forall A, A -> Type :=
  | carrier_intro : forall A x, @carrier_type A x.

(** The type [carrier_type] should not be used directly. It is used
    to implement the three following tactics:
    - [_put x] is used to return a value (leaving it on the stack)
    - [_get] is used to obtain the last returned value  
    - [_rem] is used to remove the last returned value from the 
      context. 
    The typicall usage is: [mytactic args; let result := _get in _rem; ...].
  *)

(** [_put x] adds on hypothesis of type [carrier_type x].
    [_put2 x y] and [_put3 x y z] can be used for functions
    that return pairs or triples of values. *)

Ltac _put x :=
  lets: (@carrier_intro _ x).
Ltac _put2 x y :=
  _put y; _put x.
Ltac _put3 x y z :=
  _put z; _put y; _put x.

(** [_get] returns the value [x] contained in the last hypothesis of 
    type [carrier_type x] available in the context. If fails if there
    is no such hypothesis. *)

Ltac _get :=
  match goal with H: @carrier_type _ ?t |- _ => t end.

(** [_rem] clears the last hypothesis of type [carrier_type _]. 
    If fails if there is no such hypothesis. *)

Ltac _rem :=
  match goal with H: @carrier_type _ ?t |- _ => clear H end.


(* ---------------------------------------------------------------------- *)
(** ** Testing tactics  *)

(** [show tac] executes a tactic [tac] that produces a result (not 
    performing any side-effect on the goal) and then display its result. *)

Tactic Notation "show" tactic(tac) :=
  let R := tac in pose R.

(** [dup N] produces [N] copies of the current goal. It is useful
    for building examples on which to illustrate behaviour of tactics.
    [dup] is short for [dup 2]. *)

Lemma dup_lemma : forall P, P -> P -> P.
Proof. auto. Qed.

Ltac dup_tactic N :=
  match N with
  | 0 => idtac
  | S 0 => idtac
  | S ?N' => apply dup_lemma; [ | dup_tactic N' ]
  end.

Tactic Notation "dup" constr(N) := 
  dup_tactic N.
Tactic Notation "dup" := 
  dup 2.


(* ********************************************************************** *)
(** * Tactics for development purposes *)

(* ---------------------------------------------------------------------- *)
(** ** Skipping subgoals *)

(** [skip] is an axiom that can be used to assume any goal to be true.
    It is exploited by all the tactics from this section. 
    Once all your developments are completed, you can check that this
    axiom is never used simply by changing [False] into [True] below
    (or even commenting out this entire section). *)

Variable skip : False. 

(** The [skip] tactic can be used at any time to admit the current
    goal. Using [skip] is much more efficient than using the [Focus] 
    top-level command to reach a particular subgoal. 
    Important: if you use skip a subgoal which contains uninstantiated 
    existential variables, you have to write [Admitted] and not [Qed]
    at the end of your proof. *)
    
Tactic Notation "skip" := 
  assert False; [ apply skip | contradiction ].

(** [asserts_skip H: T] adds an assumption named [H] of type [T] to the 
    current context, blindly assuming that it is true.
    [asserts_skip: T] and [skip H: T] and [skip: T] 
    are other possible syntax. *)

Tactic Notation "asserts_skip" ident(H) ":" constr(T) :=
  asserts H: T; [ skip | ]. 
Tactic Notation "asserts_skip" ":" constr(T) :=
  let H := fresh "H" in asserts_skip H: T. 
Tactic Notation "skip" ident(H) ":" constr(T) :=
  asserts_skip H: T. 
Tactic Notation "skip" ":" constr(T) :=
  asserts_skip: T. 

(** [skip_goal H] applies to any goal. It simply assumes
    the current goal to be true. The assumption is named "H".
    It is useful to set up proof by induction or coinduction.
    Syntax [skip_goal] is also accepted.*)

Tactic Notation "skip_goal" ident(H) :=
  match goal with |- ?G => asserts_skip H: G end.

Tactic Notation "skip_goal" :=
  let IH := fresh "IH" in skip_goal IH.

(** [rewrite_skip T] can be applied when [T] is an equality.
    It blindly assumes this equality to be true, and rewrite it in
    the goal. *)

Tactic Notation "rewrite_skip" constr(T) :=
  let M := fresh in asserts_skip M: T; rewrite M; clear M.

(** [rewrite_skip T in H] is similar as [rewrite_skip], except that
    it rewrites in hypothesis [H]. *)

Tactic Notation "rewrite_skip" constr(T) "in" hyp(H) :=
  let M := fresh in asserts_skip M: T; rewrite M in H; clear M.

(** [rewrites_all_skip T] is similar as [rewrite_skip], except that
    it rewrites everywhere (goal and all hypotheses). *)

Tactic Notation "rewrite_all_skip" constr(T) "in" hyp(H) :=
  let M := fresh in asserts_skip M: T; rewrite_all M; clear M.

(** [induction_skip E] applies to any goal. It simply assumes
    the current goal to be true (the assumption is named "IH" by
    default), and call [destruct E] instead of [induction E]. 
    It is useful to try and set up a proof by induction
    first, and fix the applications of the induction hypotheses
    during a second pass on the proof. *)

Tactic Notation "induction_skip" constr(E) :=
  let IH := fresh "IH" in skip_goal IH; destruct E.


(* ********************************************************************** *)
(** * Todo-list *)

(**
- a variation on [destruct] which generalize in the context all
  the variables and hypotheses that are generated, so as to allow
  one to name these hypotheses using [introv].
- a variation on [induction] which is able to deal with all the
  steps involved to work around the limitation of [induction],
  (factorizes [gen_eq; induction; intros; subst]).
- a generalization of [forward] and [apply_in] that would allow
  to provide manually several arguments at once, with an option
  to skip the non-dependent arguments.
- [forwards I1 I2 I3: E] to name conjunctions.
- [unfolds in H] should do [unfold] on head constant. 
- [fequals] should work better on n-ary tuples
*)



